import logging

from .models import Endpoint
from .client import Client

log = logging.getLogger(__name__)


def configure(name, settings, prefix='zipkin.',
              use_requests=True, use_celery=True, use_xmlrpclib=True):
    """ Include the zipkin definitions """

    endpoint = Endpoint(name)
    Client.configure(settings, prefix=prefix)

    # Install in libs here
    if use_requests:
        from .binding.requests import bind as bind_requests
        bind_requests()

    if use_celery:
        from .binding.celery import bind as bind_celery
        bind_celery(endpoint)

    if use_xmlrpclib:
        from .binding.xmlrpclib import bind as bind_xmlrpclib
        bind_xmlrpclib()

    return endpoint



# export the API here
from .api import trace, get_current_trace, stack_trace

from .config import configure
from .thread import local  # XXX remove me from here


__version__ = '0.5.3'

__all__ = ['trace', 'get_current_trace', 'stack_trace', 'configure', 'local']



import pkg_resources
import thriftpy

filename = pkg_resources.resource_filename('zipkin',
                                           'resources/zipkinCore.thrift')
zipkincore_thrift = thriftpy.load(filename, module_name='zipkincore_thrift')



from .models import Annotation
from .thread import local

__all__ = ['trace', 'get_current_trace']


def trace(name):
    """ A decorator that trace the decorated function """

    if hasattr(name, '__call__'):
        return trace(name.__name__)(name)

    def func_decorator(func):
        def wrapper(*args, **kwargs):

            try:
                try:
                    recording = True
                    trace = local().child(name)
                    annotation = Annotation.server_recv()
                    trace.record(annotation)
                except Exception:
                    recording = False
                return func(*args, **kwargs)
            finally:
                if recording:
                    trace.record(Annotation.server_send())
                    local().pop()

        return wrapper
    return func_decorator


def get_current_trace():
    return local().current


def stack_trace(trace):
    return local().append(trace)



import random
import struct
import socket

from base64 import b64encode
from six import text_type
from thriftpy.protocol import TBinaryProtocol
from thriftpy.transport import TMemoryBuffer

from .zipkin import zipkincore_thrift as ttypes


def int_or_none(val):
    if val is None:
        return None

    return int(val, 16)


def hex_str(n):
    return '%0.16x' % (n,)


def uniq_id():
    """
    Create a random 64-bit signed integer appropriate
    for use as trace and span IDs.

    XXX: By experimentation zipkin has trouble recording traces with ids
    larger than (2 ** 56) - 1

    @returns C{int}
    """
    return random.randint(0, (2 ** 56) - 1)


def base64_thrift(thrift_obj):
    trans = TMemoryBuffer()
    tbp = TBinaryProtocol(trans)

    thrift_obj.write(tbp)

    return b64encode(bytes(trans.getvalue())).strip()


def ipv4_to_int(ipv4):
    return struct.unpack('!i', socket.inet_aton(ipv4))[0]


def binary_annotation_formatter(annotation, host=None):
    annotation_types = {
        'string': ttypes.AnnotationType.STRING,
        'bytes': ttypes.AnnotationType.BYTES,
    }

    annotation_type = annotation_types[annotation.annotation_type]

    value = annotation.value

    if isinstance(value, text_type):
        value = value.encode('utf-8')

    return ttypes.BinaryAnnotation(
        annotation.name,
        value,
        annotation_type,
        host)


def base64_thrift_formatter(trace, annotations):
    thrift_annotations = []
    binary_annotations = []

    for annotation in annotations:
        host = None
        if annotation.endpoint:
            host = ttypes.Endpoint(
                ipv4=ipv4_to_int(annotation.endpoint.ip),
                port=annotation.endpoint.port,
                service_name=annotation.endpoint.service_name)

        if annotation.annotation_type == 'timestamp':
            thrift_annotations.append(ttypes.Annotation(
                timestamp=annotation.value,
                value=annotation.name,
                host=host))
        else:
            binary_annotations.append(
                binary_annotation_formatter(annotation, host))

    thrift_trace = ttypes.Span(
        trace_id=trace.trace_id,
        name=trace.name,
        id=trace.span_id,
        parent_id=trace.parent_span_id,
        annotations=thrift_annotations,
        binary_annotations=binary_annotations
    )

    return base64_thrift(thrift_trace)



import unittest
import os

import requests

from ..utils import Server
from ...binding.requests import bind as bind_requests
from ...models import Endpoint, Trace
from ...thread import local
from ...util import hex_str


def request_json(response):
    """Signature for response.json varies between
    requests versions, this wrapper finds out which api
    to use
    """
    if hasattr(response.json, '__call__'):
        return response.json()
    else:
        return response.json


class RequestsTestCase(unittest.TestCase):
    """
    This test class starts up a HttpBin instance locally for test-purpose
    """

    def __init__(self, methodName='runTest'):
        super(RequestsTestCase, self).__init__(methodName)
        bind_requests()

    def setUp(self):
        self.http_proxy = os.environ.pop('http_proxy', None)
        self.https_proxy = os.environ.pop('https_proxy', None)

        self.httpbin = Server()
        self.httpbin.start()

    def tearDown(self):
        self.httpbin.stop()
        if self.http_proxy:
            os.environ['http_proxy'] = self.http_proxy
        if self.https_proxy:
            os.environ['https_proxy'] = self.https_proxy


class RequestsNoContextTestCase(RequestsTestCase):
    def test_http_headers_no_context(self):
        resp = requests.get(self.httpbin.url + '/headers')
        headers = request_json(resp)

        self.assertEqual(headers.get('X-B3-TraceId', None), None,
                         "There should be no B3 headers if context "
                         "is not present")


class RequestsWithContextTestCase(RequestsTestCase):
    def setUp(self):
        super(RequestsWithContextTestCase, self).setUp()

        endpoint = Endpoint('127.0.0.1', 0, 'RequestsWithContextTestCase')
        self.trace = Trace('requests',
                           None, None, None,
                           endpoint=endpoint)
        local().replace(self.trace)

    def tearDown(self):
        super(RequestsWithContextTestCase, self).tearDown()
        local().reset()

    def test_http_headers_context(self):
        resp = requests.get(self.httpbin.url + '/headers')
        headers = request_json(resp)['headers']

        self.assertEqual(headers.get('X-B3-Traceid', None),
                         hex_str(self.trace.trace_id),
                         "There should be B3 headers if context is present")



from pyramid.events import NewRequest

from zipkin.config import configure
from .pyramidhook import wrap_request


def includeme(config):
    """Include the zipkin definitions"""

    settings = config.registry.settings
    if 'zipkin.collector' not in settings:
        import logging
        logging.getLogger(__name__).warn('The plugin zipkin.binding.pyramid'
                                         'is active but not configured. '
                                         'Check the doc.')
        return
    name = settings.get('zipkin.service_name', config.registry.__name__)
    endpoint = configure(name, settings)

    config.add_subscriber(wrap_request(endpoint), NewRequest)



import logging

from zipkin import local
from zipkin.models import Annotation, Trace
from zipkin.util import hex_str, int_or_none

from zipkin.client import log


endpoint = None
logger = logging.getLogger(__name__)


def task_send_handler(body, exchange, routing_key, headers, **kwargs):
    trace = local().current

    if not trace:
        logger.warn('No zipkin parent trace found, ignoring tracing task.')
        return

    forwarded_trace = trace.child_noref("subservice")
    headers['X-B3-TraceId'] = hex_str(forwarded_trace.trace_id)
    headers['X-B3-SpanId'] = hex_str(forwarded_trace.span_id)
    forwarded_trace.record(Annotation.client_send())

    if forwarded_trace.parent_span_id is not None:
        headers['X-B3-ParentSpanId'] = hex_str(forwarded_trace.parent_span_id)


def task_prerun_handler(task_id, task, **kwargs):
    request = task.request

    trace = Trace('Task execute %r' % task.name,
                  int_or_none(request.headers.get('X-B3-TraceId', None)),
                  int_or_none(request.headers.get('X-B3-SpanId', None)),
                  int_or_none(request.headers.get('X-B3-ParentSpanId', None)),
                  endpoint=endpoint)

    setattr(request, 'trace', trace)
    local().append(trace)
    trace.record(Annotation.server_recv())


def task_postrun_handler(task_id, task, **kwargs):
    trace = local().current
    trace.record(Annotation.server_send())
    trace.record(Annotation.client_recv())

    log(trace)
    local().pop()



from zipkin import local
from zipkin.models import Trace, Annotation
from zipkin.util import int_or_none
from zipkin.client import log


def wrap_request(endpoint):
    def wrap(event):
        request = event.request
        headers = request.headers
        trace = Trace(request.method + ' ' + request.path_qs,
                      int_or_none(headers.get('X-B3-TraceId', None)),
                      int_or_none(headers.get('X-B3-SpanId', None)),
                      int_or_none(headers.get('X-B3-ParentSpanId', None)),
                      endpoint=endpoint)

        setattr(request, 'trace', trace)
        local().append(trace)
        trace.record(Annotation.server_recv())
        request.add_finished_callback(log_response(endpoint))

    return wrap


def log_response(endpoint):
    def wrap(request):
        trace = request.trace
        trace.record(Annotation.server_send())

        log(trace)
        local().pop()

    return wrap



import logging

from zipkin import get_current_trace
from zipkin.models import Annotation

log = logging.getLogger(__name__)
endpoints = {}


def before_cursor_execute(conn, cursor, statement, parameters, context,
                          executemany):
    try:
        endpoint = endpoints.get(conn.engine)
        parent_trace = get_current_trace()

        if not parent_trace:
            log.error('No parent found while tracing SQL')
            return

        try:
            context.trace = parent_trace.child('SQL', endpoint=endpoint)
            abstract = context
        except AttributeError:
            cursor.trace = parent_trace.child('SQL', endpoint=endpoint)
            abstract = cursor

        abstract.trace.record(Annotation.string('db.statement', statement))

        if parameters:
            if isinstance(parameters, dict):
                parameters = dict([(key, getattr(param, 'logged_value', param))
                                   for key, param in parameters.items()])
            else:
                parameters = [getattr(param, 'logged_value', param)
                              for param in parameters]

        abstract.trace.record(Annotation.string('db.parameters',
                              repr(parameters)))
        abstract.trace.record(Annotation.string('span.kind', 'client'))
        abstract.trace.record(Annotation.server_recv())
    except Exception:
        log.exception('Unexpected exception while tracing SQL')


def after_cursor_execute(conn, cursor, statement, parameters, context,
                         executemany):
    if not hasattr(context, 'trace') and not hasattr(cursor, 'trace'):
        return
    abstract = context if hasattr(context, 'trace') else cursor

    try:
        abstract.trace.record(Annotation.string('status', 'OK'))
        abstract.trace.record(Annotation.server_send())
    except Exception:
        log.exception('Unexpected exception while tracing SQL')


def dbapi_error(conn, cursor, statement, parameters, context, exception):
    if not hasattr(context, 'trace') and not hasattr(cursor, 'trace'):
        return
    abstract = context if hasattr(context, 'trace') else cursor

    try:
        abstract.trace.record(Annotation.string('status', 'KO'))
        abstract.trace.record(Annotation.server_send())
    except Exception:
        log.exception('Unexpected exception while tracing SQL')



from zipkin import local
from zipkin.client import log
from zipkin.models import Annotation, Trace
from zipkin.util import int_or_none

from flask import request

endpoint = None


def pre_request(app, **extra):
    headers = request.headers
    trace = Trace(request.method + ' ' + request.url,
                  int_or_none(headers.get('X-B3-TraceId', None)),
                  int_or_none(headers.get('X-B3-SpanId', None)),
                  int_or_none(headers.get('X-B3-ParentSpanId', None)),
                  endpoint=endpoint)

    setattr(request, 'trace', trace)
    local().append(trace)
    trace.record(Annotation.string('http.uri', request.url))
    trace.record(Annotation.server_recv())


def pre_response(app, response, **extra):
    request.trace.record(Annotation.string('http.responsecode',
                                           '{0}'.format(response.status_code)))
    request.trace.record(Annotation.server_send())
    log(request.trace)
    local().pop()



from zipkin import local
from zipkin.models import Annotation
from zipkin.util import hex_str


def pre_request(request):
    parent_trace = local().current
    if not parent_trace:
        return request

    request.trace = parent_trace.child("requests:%s %s" %
                                       (request.method, request.url))
    forwarded_trace = request.trace.child_noref("subservice")

    request.headers['X-B3-TraceId'] = hex_str(forwarded_trace.trace_id)
    request.headers['X-B3-SpanId'] = hex_str(forwarded_trace.span_id)
    if forwarded_trace.parent_span_id is not None:
        request.headers['X-B3-ParentSpanId'] = \
            hex_str(forwarded_trace.parent_span_id)

    request.trace.record(Annotation.string('http.method', request.method))
    request.trace.record(Annotation.string('http.url', request.url))
    request.trace.record(Annotation.string('span.kind', 'client'))
    request.trace.record(Annotation.server_recv())

    return request


def pre_response(resp, req=None):
    if not req:
        req = resp.request

    if not hasattr(req, 'trace'):
        return resp

    req.trace.record(Annotation.string('http.status_code',
                                       '{0}'.format(getattr(resp, 'status',
                                                            None))))
    req.trace.record(Annotation.server_send())

    return resp


try:
    from requests.adapters import HTTPAdapter

    class ZipkinHTTPAdapter(HTTPAdapter):
        def send(self, request, stream=False, timeout=None, verify=True,
                 cert=None, proxies=None):
            pre_request(request)
            return super(ZipkinHTTPAdapter, self).send(request, stream,
                                                       timeout, verify, cert,
                                                       proxies)

        def build_response(self, req, resp):
            pre_response(resp, req)

            return super(ZipkinHTTPAdapter, self).build_response(req, resp)

except ImportError:
    # requests < 1.0.0
    pass


def session_init(init):
    def func(self, *args, **kwargs):
        init(self, *args, **kwargs)
        if hasattr(self, 'mount'):
            self.mount('http://', ZipkinHTTPAdapter())
            self.mount('https://', ZipkinHTTPAdapter())
        else:
            self.hooks['pre_request'] = pre_request
            self.hooks['response'] = pre_response
    return func



import os
import re
import sys
from setuptools import setup, find_packages

name = 'zipkin'
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.rst')) as f:
    CHANGES = f.read()
with open(os.path.join(here, name, '__init__.py')) as version:
    VERSION = re.compile(r".*__version__ = '(.*?)'",
                         re.S).match(version.read()).group(1)

requires = [
    'thriftpy',
    'six',
    ]

tests_require = [
    'mock'
    ]

if sys.version_info[:2] < (2, 7):
    tests_require.append('unittest2')

extras_require = {
    'pyramid': [
        'pyramid',
    ],
    'celery': [
        'celery >= 3.1',
    ],
    'requests': [
        'requests',
    ],
    'flask': [
        'flask', 'blinker',
    ],
    'dev': ['pyramid', 'celery', 'requests', 'flask', 'sphinx'],
    'test': tests_require,
}

setup(name='zk',
      version=VERSION,
      description='zipkin',
      long_description=README + '\n\n' + CHANGES,
      classifiers=['Programming Language :: Python',
                   ],
      author='Gandi',
      author_email='feedback@gandi.net',
      url='',
      keywords='zipkin metrics trace',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      extras_require=extras_require,
      tests_require=tests_require,
      test_suite='{name}.tests'.format(name=name),
      )
